/* app/builder/KMBuildInit.cpp
 *
 * Copyright (C) 2023 OpenKylin, Inc
 */

#include "KMBuildInit.h"

#include <fstream>
#include <nlohmann/json.hpp>

#include "common/KMLogger.h"
#include "common/KMOptions.h"
#include "common/kmtranslation.h"
#include "common/KMException.h"
#include "common/KMInstalledAppQuery.h"
#include "common/KMInfoJson.h"

class KMBuildInit::Options : public KMOptions
{
public:
    Options() = default;
    virtual ~Options() = default;

protected:
    void preParseHook() override;
    void postParseHook() override;

private:
    void addOptions();
    void checkOptions(const opt::variables_map &varsMap);

public:
    std::string m_arch;
    std::string m_runtime;
    std::string m_base;
    std::string m_type;
    std::string m_channel;
    bool m_help = false;

    std::string m_buildDir;
    std::string m_appid;

private:
    std::set<std::string> m_setTypes{ "app", "runtime", "extension" };
};

void KMBuildInit::Options::preParseHook()
{
    addOptions();
}

void KMBuildInit::Options::postParseHook()
{
    if (m_help)
    {
        showUsage();
        exit(EXIT_SUCCESS);
    }

    if (m_buildDir.empty() || m_appid.empty())
    {
        KMError(_("DIRECTORY APPID must be specified"));
        showUsage();
        exit(EXIT_FAILURE);
    }

    if (m_setTypes.end() == m_setTypes.find(m_type))
    {
        KMError(m_type + _(" is not a valid build type name, use app, runtime or extension"));
        exit(EXIT_FAILURE);
    }

    std::string errMsg;
    if (!KMBuildinsUtils::isValidAppName(m_appid, errMsg))
    {
        KMError(errMsg);
        exit(EXIT_FAILURE);
    }
}

void KMBuildInit::Options::addOptions()
{
    m_optionsDescription = std::make_unique<opt::options_description>(_("\nUsage:\n \tkaiming build-init [OPTIONS] DIRECTORY APPID - Initialize a directory for building "
                                                                        "\nExample:\n \tkaiming build-init --base=top.openkylin.base/stable/2.0.0 --runtime=top.openkylin.runtime/stable/2.0.0 build top.openkylin.demo"
                                                                        "\n\nOptions"));

    opt::options_description helpOptions(_("Help Options"));
    helpOptions.add_options()("help,h", opt::bool_switch(&m_help), _("Show help options"));
    opt::options_description appOptions(_("Application Options"));
    appOptions.add_options()("arch", opt::value<std::string>(&m_arch)->default_value(KMUtil::kmGetArch()), _("Arch to use"))
                            ("runtime", opt::value<std::string>(&m_runtime), _("Runtime info, format : ${runtime-id}/${runtime-channel}/${runtime-version}, such as : top.openkylin.runtime/stable/2.0.0"))
                            ("base", opt::value<std::string>(&m_base), _("Base info, format : ${base-id}/${base-channel}/${base-version}, such as : top.openkylin.base/stable/2.0.0"))
                            ("kind", opt::value<std::string>(&m_type)->default_value("app"), _("Specify the build type (app, runtime, extension)"))
                            ("channel", opt::value<std::string>(&m_channel)->default_value("stable"), _("Release channel, such as (main, stable, etc.)"));

    opt::options_description positionalOptions(_("Positional Options"));
    positionalOptions.add_options()("DIRECTORY", opt::value<std::string>(&m_buildDir), _("The directory for building"));
    positionalOptions.add_options()("APPID", opt::value<std::string>(&m_appid), _("Appid"));

    m_optionsDescription->add(helpOptions).add(appOptions).add(positionalOptions);

    m_optionsPositionalDescription->add("DIRECTORY", 1);
    m_optionsPositionalDescription->add("APPID", 1);
}

class KMBuildInit::Private
{
public:
    std::unique_ptr<Options> m_kmOptions;

    KMBuildType m_type;
    KMStorageDir m_baseDir;
    KMStorageDir m_runtimeDir;
    std::optional<KMRef> m_originalBaseRef;
    std::optional<KMRef> m_originalRuntimeRef;
    KMRef m_baseRef;
    KMRef m_runtimeRef;

    std::string m_filesDir;
    std::string m_varDir;
    std::string m_varTmpDir;
    std::string m_varRunDir;
    std::string m_metadataFile;
    std::string m_usrDir;
    std::string m_varDeployFiles;
};

REGISTER_SUBCOMMAND_DYNCREATE(build-init, KMBuildInit)

void KMBuildInit::checkBase()
{
    KMInfoJson::InfoBase appBase, runtimeBase;
    fs::path runtimePathBoot = KMStorageDir::getSystemDefaultBaseDirLocation();

    //加载baseId与baseVersion  /opt/kaiming/layers/${channel}/${arch}/${kind}/${id}/{module}/${version}/info.json
    fs::path infoPath = runtimePathBoot / d->m_runtimeRef.channel / d->m_runtimeRef.arch /d->m_runtimeRef.kind / d->m_runtimeRef.id / d->m_runtimeRef.module / d->m_runtimeRef.version /"info.json"; 

    KMInfoJson infoJson;
    infoJson.loadFile(infoPath);

    runtimeBase.m_baseString = infoJson.base;
    runtimeBase.baseParse();

    //解析appBase
    appBase.m_baseId = d->m_baseRef.id;
    appBase.m_baseVersion= d->m_baseRef.version;
    
    if (appBase.m_baseId != runtimeBase.m_baseId)
    {
        throw KMException("The appBaseId '" + appBase.m_baseId + "' and the runtimeBaseId '" + runtimeBase.m_baseId + "' is not matched" );
    }

    //版本号解析,判断version的x.y部分
    
    appBase.versionParse();
    runtimeBase.versionParse();

    if(! appBase.checkVersion(runtimeBase))
    {
        throw KMException("The appBaseVersion '" + appBase.m_baseVersion + "' and the runtimeVersion '" + runtimeBase.m_baseVersion + "' is not matched" );
    }

}
KMBuildInit::KMBuildInit()
    : d(std::make_unique<Private>())
{
    d->m_kmOptions = std::make_unique<Options>();

    enableVfs();
}

KMBuildInit::~KMBuildInit() { }

int KMBuildInit::despose(int argc, char **argv)
{
    KMTrace("KMBuildInit::despose invoke begin");

    init(argc, argv);

    int ret = run();

    KMTrace("KMBuildInit::despose invoke end");
    return ret;
}

void KMBuildInit::init(int argc, char **argv)
{
    KMTrace("KMBuildInit::init invoke begin");

    d->m_kmOptions->parseCommandLine(argc, argv);

    // 当前首先支持构建app；后续要支持构建扩展、运行时，设计上先预留
    if (d->m_kmOptions->m_type == "extension")
    {
        d->m_type = KMBUILDTYPE_EXTERNSION;
    }
    else if (d->m_kmOptions->m_type == "runtime")
    {
        d->m_type = KMBUILDTYPE_RUNTIME;
    }
    else
    {
        d->m_type = KMBUILDTYPE_APP;
    }

    // 解析base、runtime，并校验是否已经安装
    d->m_originalBaseRef = KMBuildInit::parsePref(d->m_kmOptions->m_base);
    if (!d->m_originalBaseRef)
    {
        throw KMException("The config of base '" + d->m_kmOptions->m_base + "' is wrong");
    }

    // 校验base是否已安装
    KMInstalledAppQuery query;
    std::vector<KMRef> bases = query.channel(d->m_originalBaseRef->channel)
                                    .arch(d->m_kmOptions->m_arch)
                                    .kind(BASE_TYPE_BASE)
                                    .id(d->m_originalBaseRef->id)
                                    .module(MODULE_NAME_DEV)
                                    .version(d->m_originalBaseRef->version)
                                    .query();
    if (bases.empty())
    {
        throw KMException("No matching development base found, please install it : kaiming install " 
                        + KMUtil::buildFilename(d->m_originalBaseRef->id, d->m_originalBaseRef->version, MODULE_NAME_DEV, d->m_originalBaseRef->channel));
    }
    d->m_baseRef = bases.at(0);

    // 解析runtime，并校验是否已经安装
    if (!d->m_kmOptions->m_runtime.empty())
    {
        d->m_originalRuntimeRef = KMBuildInit::parsePref(d->m_kmOptions->m_runtime);
        if (!d->m_originalRuntimeRef)
        {
            throw KMException("The config of runtime '" + d->m_kmOptions->m_runtime + "' is wrong");
        }

        std::vector<KMRef> runtimes = query.channel(d->m_originalRuntimeRef->channel)
                                           .arch(d->m_kmOptions->m_arch)
                                           .kind(BASE_TYPE_RUNTIME)
                                           .id(d->m_originalRuntimeRef->id)
                                           .module(MODULE_NAME_DEV)
                                           .version(d->m_originalRuntimeRef->version)
                                           .query();
        if (runtimes.empty())
        {
            throw KMException("No matching development runtime found, please install it : kaiming install " 
                            + KMUtil::buildFilename(d->m_originalRuntimeRef->id, d->m_originalRuntimeRef->version, MODULE_NAME_DEV, d->m_originalRuntimeRef->channel));
        }
        d->m_runtimeRef = runtimes.at(0);

        //  app的base和runtime的base要保持一致
        checkBase();
       
    }

    d->m_filesDir = d->m_kmOptions->m_buildDir + "/files";
    d->m_varDir = d->m_kmOptions->m_buildDir + "/var";
    d->m_varTmpDir = d->m_kmOptions->m_buildDir + "/var/tmp";
    d->m_varRunDir = d->m_kmOptions->m_buildDir + "/var/run";
    d->m_metadataFile = d->m_kmOptions->m_buildDir + "/info.json";

    if (KMUtil::pathExists(d->m_filesDir))
    {
        throw KMException(std::string(_("Build directory already initialized : ")) + d->m_kmOptions->m_buildDir);
    }

    KMTrace("KMBuildInit::init invoke end");
}

int KMBuildInit::run()
{
    KMTrace("KMBuildInit::run invoke begin");

    // 创建构建目录
    std::error_code ec;
    if (!KMUtil::mkpath(d->m_kmOptions->m_buildDir, ec))
    {
        throw KMException(ec.message());
    }

    // 创建files目录
    if (!KMUtil::mkpath(d->m_filesDir, ec))
    {
        throw KMException(ec.message());
    }

    // 创建/var/tmp目录
    if (!KMUtil::mkpath(d->m_varTmpDir, ec))
    {
        throw KMException(ec.message());
    }

    // 创建/var/run软连接
    if (!KMUtil::createDirectorySymlink("/run", d->m_varRunDir, ec))
    {
        throw KMException(ec.message());
    }

    // 生成info.json文件
    generateMetadata();

    KMTrace("KMBuildInit::run invoke end");
    return EXIT_SUCCESS;
}

/**
     * @brief : 解析base/runtime等的pref
     * @param : [in] pref, 格式如：top.openkylin.base/stable/2.0.0
     * @return: std::optional<KMRef>
 */ 
std::optional<KMRef> KMBuildInit::parsePref(const std::string& pref)
{
    std::optional<KMRef> ref;

    std::vector<std::string> splits = KMUtil::splitString(pref, "/");
    if (splits.size() != 3)
    {
        KMError("Format such as : ${id}/${channel}/${version}");
        return ref;
    }

    std::string errMsg;
    if (!KMBuildinsUtils::isValidAppName(splits.at(0), errMsg))
    {
        KMError(errMsg);
        return ref;
    }

    ref = {
        .id = splits.at(0),
        .channel = splits.at(1),
        .version = splits.at(2),
        .arch = "",
        .module = "",
        .branch = "",
        .kind = "",
        .refCommitId = ""
    };

    return ref;
}

void KMBuildInit::generateMetadata()
{
    KMTrace("KMBuildInit::generateMetadata invoke begin");

    KMInfoJson doc;
    
    doc.id = d->m_kmOptions->m_appid;
    doc.archs.push_back(d->m_kmOptions->m_arch);
    doc.kind = d->m_kmOptions->m_type;
    doc.channel = d->m_kmOptions->m_channel;
    doc.base = d->m_originalBaseRef->channel + ":" + d->m_originalBaseRef->id + "/" + d->m_originalBaseRef->version + "/" + d->m_kmOptions->m_arch;
    if (d->m_originalRuntimeRef)
    {
        doc.runtime = d->m_originalRuntimeRef->channel + ":" + d->m_originalRuntimeRef->id + "/" + d->m_originalRuntimeRef->version + "/" + d->m_kmOptions->m_arch;
    }
    else
    {
        doc.runtime = "";
    }
    doc.saveFile(KMUtil::buildFilename(d->m_kmOptions->m_buildDir, "info.json"));

    KMTrace("KMBuildInit::generateMetadata invoke end");

}

///////////////////////////old begin////////////////////////////////

// #define GIT_IGNORE_FILE \
//   "# This file is autogenerated by kaiming build-init\n" \
//   "*\n"

// class KMBuildInit::Options : public KMOptions
// {
// public:
//     Options() = default;
//     virtual ~Options() = default;

// protected:
//     void preParseHook() override;
//     void postParseHook() override;

// private:
//     void addOptions();
//     void checkOptions(const opt::variables_map &varsMap);

// public:
//     std::string m_arch;
//     std::string m_var;
//     std::string m_base;
//     std::string m_baseVersion;
//     std::vector<std::string> m_baseExtensions;
//     std::string m_extensionTag;
//     std::string m_type;
//     std::vector<std::string> m_tags;
//     std::vector<std::string> m_sdkExtensions;
//     std::vector<std::string> m_extensions;
//     std::string m_sdkDir;
//     bool m_writableSdk = false;
//     bool m_update = false;
//     bool m_help = false;

//     std::string m_buildDir;
//     std::string m_appid;
//     std::string m_sdk;
//     std::string m_runtime;
//     std::string m_branch;

// private:
//     std::set<std::string> m_setTypes{ "app", "runtime", "extension" };
// };

// void KMBuildInit::Options::preParseHook()
// {
//     addOptions();
// }

// void KMBuildInit::Options::postParseHook()
// {
//     checkOptions(m_varsMap);

//     if (m_help)
//     {
//         showUsage();
//         exit(EXIT_SUCCESS);
//     }

//     if (m_buildDir.empty() || m_appid.empty() || m_sdk.empty() || m_runtime.empty())
//     {
//         KMError(_("DIRECTORY APPNAME SDK RUNTIME must be specified"));
//         showUsage();
//         exit(EXIT_FAILURE);
//     }

//     if (m_setTypes.end() == m_setTypes.find(m_type))
//     {
//         KMError(m_type + _(" is not a valid build type name, use app, runtime or extension"));
//         exit(EXIT_FAILURE);
//     }

//     std::string errMsg;
//     if (!KMBuildinsUtils::isValidAppName(m_appid, errMsg))
//     {
//         KMError(errMsg);
//         exit(EXIT_FAILURE);
//     }
// }

// void KMBuildInit::Options::addOptions()
// {
//     m_optionsDescription = std::make_unique<opt::options_description>(_("\nUsage:\n \tkaiming build-init [OPTION...] DIRECTORY APPNAME SDK RUNTIME [BRANCH] - Initialize a directory for building "
//                                                                         "\nExample:\n \tkaiming build-init --arch=x86_64 /home/kylin/qtdemo/ top.openkylin.qtdemo top.openkylin.Sdk top.openkylin.Platform 5.12-23.12"
//                                                                         "\n\nOptions"));

//     opt::options_description helpOptions(_("Help Options"));
//     helpOptions.add_options()("help,h", _("Show help options"));
//     opt::options_description appOptions(_("Application Options"));
//     appOptions.add_options()("arch", opt::value<std::string>(&m_arch), _("Arch to use"))
//                             ("var,v", opt::value<std::string>(&m_var), _("Initialize var from named runtime"))
//                             ("base", opt::value<std::string>(&m_base), _("Initialize apps from named app"))
//                             ("base-version", opt::value<std::string>(&m_baseVersion), _("Specify version for --base"))
//                             ("base-extension", opt::value<std::vector<std::string>>(&m_baseExtensions), _("Include this base extension"))
//                             ("extension-tag", opt::value<std::string>(&m_extensionTag), _("Extension tag to use if building extension"))
//                             ("writable-sdk,w", _("Initialize /usr with a writable copy of the sdk"))
//                             ("type", opt::value<std::string>(&m_type)->default_value("app"), _("Specify the build type (app, runtime, extension)"))
//                             ("tag", opt::value<std::vector<std::string>>(&m_tags), _("Add a tag"))
//                             ("sdk-extension", opt::value<std::vector<std::string>>(&m_sdkExtensions), _("Include this sdk extension in /usr"))
//                             ("extension", opt::value<std::vector<std::string>>(&m_extensions), _("Add extension point info (NAME=VARIABLE[=VALUE])"))
//                             ("sdk-dir", opt::value<std::string>(&m_sdkDir), _("Where to store sdk (defaults to 'usr')"))
//                             ("update", _("Re-initialize the sdk/var"));

//     opt::options_description positionalOptions(_("Positional Options"));
//     positionalOptions.add_options()("DIRECTORY", opt::value<std::string>(&m_buildDir), _("The directory for building"));
//     positionalOptions.add_options()("APPNAME", opt::value<std::string>(&m_appid), _("Appid"));
//     positionalOptions.add_options()("SDK", opt::value<std::string>(&m_sdk), _("Sdk"));
//     positionalOptions.add_options()("RUNTIME", opt::value<std::string>(&m_runtime), _("Runtime"));
//     positionalOptions.add_options()("BRANCH", opt::value<std::string>(&m_branch), _("Brunch of sdk and runtime"));

//     m_optionsDescription->add(helpOptions).add(appOptions).add(positionalOptions);

//     m_optionsPositionalDescription->add("DIRECTORY", 1);
//     m_optionsPositionalDescription->add("APPNAME", 1);
//     m_optionsPositionalDescription->add("SDK", 1);
//     m_optionsPositionalDescription->add("RUNTIME", 1);
//     m_optionsPositionalDescription->add("BRANCH", 1);
// }

// void KMBuildInit::Options::checkOptions(const opt::variables_map &varsMap)
// {
//     checkOption(varsMap, "help", "h", m_help);
//     checkOption(varsMap, "writable-sdk", "w", m_writableSdk);
//     checkOption(varsMap, "update", nullptr, m_update);
// }

// class KMBuildInit::Private
// {
// public:
//     std::unique_ptr<Options> m_kmOptions;

//     KMBuildType m_type;
//     KMRef m_sdkRef;
//     KMStorageDir m_sdkDir;
//     KMRef m_runtimeRef;
//     KMStorageDir m_runtimeDir;
//     std::string m_extensionRuntimePref;
//     KMRef m_baseRef;

//     std::string m_filesDir;
//     std::string m_varDir;
//     std::string m_varTmpDir;
//     std::string m_varRunDir;
//     std::string m_metadataFile;
//     std::string m_gitignoreFile;
//     std::string m_usrDir;
//     std::string m_varDeployFiles;
// };

// REGISTER_SUBCOMMAND_DYNCREATE(build-init, KMBuildInit)

// KMBuildInit::KMBuildInit()
//     : d(std::make_unique<Private>())
// {
//     d->m_kmOptions = std::make_unique<Options>();

//     enableVfs();
// }

// KMBuildInit::~KMBuildInit() { }

// int KMBuildInit::despose(int argc, char **argv)
// {
//     KMTrace("KMBuildInit::despose invoke begin");

//     init(argc, argv);

//     int ret = run();

//     KMTrace("KMBuildInit::despose invoke end");
//     return ret;
// }

// void KMBuildInit::init(int argc, char **argv)
// {
//     KMTrace("KMBuildInit::init invoke begin");

//     d->m_kmOptions->parseCommandLine(argc, argv);

//     if (d->m_kmOptions->m_type == "extension")
//     {
//         d->m_type = KMBUILDTYPE_EXTERNSION;
//     }
//     else if (d->m_kmOptions->m_type == "runtime")
//     {
//         d->m_type = KMBUILDTYPE_RUNTIME;
//     }
//     else
//     {
//         d->m_type = KMBUILDTYPE_APP;
//     }

//     KaimingKinds kinds = KaimingKinds::KAIMING_KINDS_RUNTIME;
//     std::string errMsg;
//     d->m_sdkDir = KMBuildinsUtils::findInstalledPref(d->m_kmOptions->m_sdk, kinds, d->m_kmOptions->m_arch, d->m_kmOptions->m_branch, d->m_sdkRef, errMsg);
//     if (!d->m_sdkDir)
//     {
//         KMError(std::string(_("SDK not installed. ")) + errMsg);
//         exit(EXIT_FAILURE);
//     }

//     kinds = KAIMING_KINDS_RUNTIME;
//     if (d->m_type == KMBUILDTYPE_EXTERNSION)
//     {
//         kinds = KAIMING_KINDS_ALL;
//     }
//     d->m_runtimeDir = KMBuildinsUtils::findInstalledPref(d->m_kmOptions->m_runtime, kinds, d->m_kmOptions->m_arch, d->m_kmOptions->m_branch, d->m_runtimeRef, errMsg);
//     if (!d->m_runtimeDir)
//     {
//         KMError(std::string(_("RUNTIME not installed. ")) + errMsg);
//         exit(EXIT_FAILURE);
//     }

//     if (d->m_type == KMBUILDTYPE_EXTERNSION)
//     {
//         /* The "runtime" can be an app in case we're building an extension */
//         if (d->m_runtimeRef.kind == BASE_TYPE_APP)
//         {
//             std::shared_ptr<KMDeploy> deploy = d->m_runtimeDir.loadDeployed(d->m_runtimeRef);
//             if (!deploy)
//             {
//                 KMError("Can not find the runtime of the externsion!");
//                 exit(EXIT_FAILURE);
//             }

//             d->m_extensionRuntimePref = deploy->m_metadata->getRuntime();
//         }
//         else
//         {
//             d->m_extensionRuntimePref = KMUtil::dupRef(d->m_runtimeRef);
//         }
//     }

//     d->m_filesDir = d->m_kmOptions->m_buildDir + "/files";
//     d->m_varDir = d->m_kmOptions->m_buildDir + "/var";
//     d->m_varTmpDir = d->m_kmOptions->m_buildDir + "/var/tmp";
//     d->m_varRunDir = d->m_kmOptions->m_buildDir + "/var/run";
//     d->m_metadataFile = d->m_kmOptions->m_buildDir + "/metadata";
//     d->m_gitignoreFile = d->m_kmOptions->m_buildDir + "/.gitignore";

//     if (!d->m_kmOptions->m_update && KMUtil::pathExists(d->m_filesDir))
//     {
//         KMError(std::string(_("Build directory already initialized : ")) + d->m_kmOptions->m_buildDir);
//         exit(EXIT_FAILURE);
//     }

//     KMTrace("KMBuildInit::init invoke end");
// }

// int KMBuildInit::run()
// {
//     KMTrace("KMBuildInit::run invoke begin");

//     // 创建构建目录
//     std::error_code ec;
//     if (!KMUtil::mkpath(d->m_kmOptions->m_buildDir, ec))
//     {
//         KMError(std::string(d->m_kmOptions->m_buildDir + _(" create failed, error: ")) + ec.message());
//         return EXIT_FAILURE;
//     }

//     // 拷贝运行时副本到构建目录
//     std::shared_ptr<KMDeploy> sdkDeploy = d->m_sdkDir.loadDeployed(d->m_sdkRef);
//     if (!sdkDeploy)
//     {
//         KMError(_("Can not find sdk deploy!"));
//         return EXIT_FAILURE;
//     }
//     if (!copyUsrToBuildDir())
//     {
//         return EXIT_FAILURE;
//     }

//     if (!d->m_kmOptions->m_sdkExtensions.empty())
//     {
//         KMTrace("Deal sdkExtensions");
//         if (!ensureExtensions(sdkDeploy, d->m_sdkRef.arch, d->m_sdkRef.branch, d->m_kmOptions->m_sdkExtensions, d->m_usrDir))
//         {
//             KMTrace("KMBuildInit::run invoke end : ensureExtensions sdkExtensions false");
//             return EXIT_FAILURE;
//         }
//     }

//     if (!d->m_kmOptions->m_var.empty())
//     {
//         KMRef var_ref = KMUtil::constructRef(KaimingKinds::KAIMING_KINDS_RUNTIME, d->m_kmOptions->m_var, d->m_kmOptions->m_arch, d->m_kmOptions->m_branch);
//         std::string var_deploy = KMStorageDir::getDeployedDir(var_ref);
//         if (var_deploy.empty())
//         {
//             KMError(_("Can not find var deploy!"));
//             return EXIT_FAILURE;
//         }
//         d->m_varDeployFiles = var_deploy + "/files";
//     }

//     if (d->m_kmOptions->m_update)
//     {
//         KMTrace("KMBuildInit::run invoke end : update");
//         return EXIT_SUCCESS;
//     }

//     // 创建files目录
//     if (!KMUtil::mkpath(d->m_filesDir, ec))
//     {
//         KMError(std::string(d->m_filesDir + _(" create failed, error: ")) + ec.message());
//         return EXIT_FAILURE;
//     }

//     // base处理
//     if (!copyBaseToBuildDir())
//     {
//         return EXIT_FAILURE;
//     }

//     if (d->m_varDeployFiles.empty())
//     {
//         // 创建/var目录
//         if (!KMUtil::mkpath(d->m_varDir, ec))
//         {
//             KMError(std::string(d->m_varDir + _(" create failed, error: ")) + ec.message());
//             return EXIT_FAILURE;
//         }
//     }
//     else
//     {
//         KMTrace("Copy var files to /var");
//         if (!KMUtil::copyAll(d->m_varDeployFiles, d->m_varDir, ec))
//         {
//             KMError(ec.message());
//             KMTrace("KMBuildInit::run invoke end : copy var files failed!");
//             return false;
//         }
//     }

//     // 创建/var/tmp目录
//     if (!KMUtil::mkpath(d->m_varTmpDir, ec))
//     {
//         KMError(std::string(d->m_varTmpDir + _(" create failed, error: ")) + ec.message());
//         return EXIT_FAILURE;
//     }

//     // 创建/var/run软连接
//     if (!KMUtil::createDirectorySymlink("/run", d->m_varRunDir, ec))
//     {
//         KMError(_("Create /var/run symlink failed, error: ") + ec.message());
//         return EXIT_FAILURE;
//     }

//     if (!generateMetadata())
//     {
//         KMError(_("Create metadata file failed!"));
//         return EXIT_FAILURE;
//     }

//     initGitignoreFile();

//     KMTrace("KMBuildInit::run invoke end");
//     return EXIT_SUCCESS;
// }


// bool KMBuildInit::copyUsrToBuildDir()
// {
//     KMTrace("KMBuildInit::copyUsrToBuildDir invoke begin");

//     if (d->m_kmOptions->m_writableSdk || d->m_type == KMBUILDTYPE_RUNTIME)
//     {
//         d->m_usrDir = d->m_kmOptions->m_buildDir + "/usr";
//         if (!d->m_kmOptions->m_sdkDir.empty())
//         {
//             d->m_usrDir = d->m_kmOptions->m_buildDir + "/" + d->m_kmOptions->m_sdkDir;
//         }

//         std::error_code ec;
//         if (!KMUtil::removeAll(d->m_usrDir, ec))
//         {
//             KMError(ec.message());
//             return false;
//         }

//         std::string sdkDeployFiles = KMStorageDir::getRefBaseDirLocation(d->m_sdkRef) + "/files";
//         if (!KMUtil::copyAll(sdkDeployFiles, d->m_usrDir, ec))
//         {
//             KMError(ec.message());
//             return false;
//         }
//     }

//     KMTrace("KMBuildInit::copyUsrToBuildDir invoke end");
//     return true;
// }

// bool KMBuildInit::ensureExtensions(std::shared_ptr<KMDeploy> sdkDeploy, const std::string defaultArch, const std::string &defaultBranch, const std::vector<std::string> &srcExtensions, const std::string &topDir)
// {
//     KMTrace("KMBuildInit::ensureExtensions invoke begin");

//     if (!sdkDeploy)
//     {
//         return false;
//     }

//     std::shared_ptr<KMMetaData> metadata = sdkDeploy->m_metadata;
//     const std::vector<KMExtension> &extensions = KMBuildinsUtils::listExtensions(metadata, defaultArch, defaultBranch);

//     for (std::string extensionNameWithTag : srcExtensions)
//     {
//         bool find = false;

//         std::string extensionName;
//         std::string tag;
//         KMBuildinsUtils::parseExtensionWithTag(extensionNameWithTag, extensionName, tag);
//         for (KMExtension extension : extensions)
//         {
//             if (extension.name == extensionName)
//             {
//                 if (!extension.isUnmaintained)
//                 {
//                     std::string deploy = KMStorageDir::getDeployedDir(extension.ref);
//                     if (deploy.empty())
//                     {
//                         KMError(_("Requested extension ") + extension.name + "/" + defaultArch + "/" + defaultBranch + _("is not installed!"));
//                         return false;
//                     }
//                 }

//                 if (!topDir.empty())
//                 {
//                     std::string target = KMUtil::buildFilename(topDir, extension.directory);

//                     std::error_code ec;
//                     // fs::path targetParent = fs::path(KMUtil::realpath(target));
//                     // if (!KMUtil::mkpath(targetParent.string(), ec)) {
//                     //     KMError(ec.message());
//                     //     return false;
//                     // }
//                     if (!KMUtil::removeAll(target, ec))
//                     {
//                         KMError(ec.message());
//                         return false;
//                     }

//                     std::string extDeployFiles = extension.extensionFilesPath;
//                     if (!KMUtil::copyAll(extDeployFiles, target, ec))
//                     {
//                         KMError(_("Copy extension ") + extension.name + "/" + defaultArch + "/" + defaultBranch + _("files failed!"));
//                         KMError(ec.message());
//                         return false;
//                     }
//                 }
//             }
//             find = true;
//         }

//         if (!find)
//         {
//             KMError(_("Requested extension ") + extensionName + "/" + defaultArch + "/" + defaultBranch + _("is not installed!"));
//             return false;
//         }
//     }

//     KMTrace("KMBuildInit::ensureExtensions invoke end");
//     return true;
// }

// bool KMBuildInit::copyBaseToBuildDir()
// {
//     KMTrace("KMBuildInit::copyBaseToBuildDir invoke begin");

//     if (!d->m_kmOptions->m_base.empty())
//     {
//         KMTrace("Deal base begin");
//         d->m_baseRef = KMUtil::constructRef(KaimingKinds::KAIMING_KINDS_APP, d->m_kmOptions->m_base, d->m_kmOptions->m_arch, d->m_kmOptions->m_branch);
//         std::string baseDeployDir = KMStorageDir::getDeployedDir(d->m_baseRef);
//         if (baseDeployDir.empty())
//         {
//             KMError(_("Can not find base deploy!"));
//             return false;
//         }
//         std::string baseDeployFiles = baseDeployDir + "/files";
//         std::error_code ec;
//         if (!KMUtil::copyAll(baseDeployFiles, d->m_filesDir, ec))
//         {
//             KMError(ec.message());
//             KMTrace("KMBuildInit::run invoke end : copy base files failed!");
//             return false;
//         }

//         if (!d->m_kmOptions->m_baseExtensions.empty())
//         {
//             std::shared_ptr<KMDeploy> baseDeploy = d->m_sdkDir.loadDeployed(d->m_baseRef);
//             if (!baseDeploy)
//             {
//                 KMError(_("Can not find sdk deploy!"));
//                 return false;
//             }

//             if (!ensureExtensions(baseDeploy, d->m_baseRef.arch, d->m_baseRef.branch, d->m_kmOptions->m_baseExtensions, d->m_filesDir))
//             {
//                 KMTrace("KMBuildInit::run invoke end : ensureExtensions baseExtensions false");
//                 return false;
//             }
//         }
//     }

//     KMTrace("KMBuildInit::copyBaseToBuildDir invoke end");
//     return true;
// }

// bool KMBuildInit::generateMetadata()
// {
//     KMTrace("KMBuildInit::generateMetadata invoke begin");

//     KMMetaData metadata(d->m_type == KMBUILDTYPE_APP);

//     // [Application] or [Runtime]
//     std::string appGroup = metadata.isApp() ? METADATA_GROUP_APPLICATION : METADATA_GROUP_RUNTIME;
//     metadata.setStringValue(appGroup, METADATA_KEY_NAME, d->m_kmOptions->m_appid);
//     /* The "runtime" can be an app in case we're building an extension */
//     if (d->m_runtimeRef.kind == BASE_TYPE_RUNTIME)
//     {
//         metadata.setStringValue(appGroup, METADATA_KEY_RUNTIME, KMUtil::dupPref(d->m_runtimeRef));
//     }
//     if (d->m_sdkRef.kind == BASE_TYPE_RUNTIME)
//     {
//         metadata.setStringValue(appGroup, METADATA_KEY_SDK, KMUtil::dupPref(d->m_sdkRef));
//     }
//     if (!d->m_baseRef.id.empty())
//     {
//         metadata.setStringValue(appGroup, METADATA_KEY_BASE, KMUtil::dupRef(d->m_baseRef));
//     }
//     if (!d->m_kmOptions->m_tags.empty())
//     {
//         metadata.setStringListValue(appGroup, METADATA_KEY_TAGS, d->m_kmOptions->m_tags);
//     }

//     if (d->m_type == KMBUILDTYPE_EXTERNSION)
//     {
//         metadata.setStringValue(METADATA_GROUP_EXTENSION_OF, METADATA_KEY_REF, KMUtil::dupPref(d->m_runtimeRef));
//         metadata.setStringValue(METADATA_GROUP_EXTENSION_OF, METADATA_KEY_RUNTIME, d->m_extensionRuntimePref);
//         metadata.setStringValue(METADATA_GROUP_EXTENSION_OF, METADATA_KEY_TAG, d->m_kmOptions->m_extensionTag);
//     }

//     for (const std::string &extension : d->m_kmOptions->m_extensions)
//     {
//         std::vector<std::string> elements = KMUtil::splitString(extension, "=");
//         if (elements.size() < 2)
//         {
//             KMError(_("Too few elements in --extension argument, format should be NAME=VAR[=VALUE]"));
//             return false;
//         }
//         std::string errMsg;
//         if (!KMBuildinsUtils::isValidAppName(elements.at(0), errMsg))
//         {
//             KMError(_("Invalid extension name ") + elements.at(0));
//             KMError(errMsg);
//             return false;
//         }

//         std::string groupName(METADATA_GROUP_PREFIX_EXTENSION);
//         groupName += elements.at(0);
//         std::string key = elements.at(1);
//         std::string value("true");
//         if (elements.size() > 2)
//         {
//             value = elements.at(2);
//         }
//         metadata.setStringValue(groupName, key, value);
//     }

//     if (!metadata.saveFile(d->m_metadataFile))
//     {
//         KMError(_("Save metadata file failed!"));
//         return false;
//     }

//     KMTrace("KMBuildInit::generateMetadata invoke end");
//     return true;
// }

// bool KMBuildInit::initGitignoreFile()
// {
//     KMTrace("KMBuildInit::initGitignoreFile invoke begin");

//     std::fstream gitignore(d->m_gitignoreFile, std::ios::out | std::ios::trunc);
//     gitignore << GIT_IGNORE_FILE;
//     gitignore.flush();

//     KMTrace("KMBuildInit::initGitignoreFile invoke end");
//     return true;
// }
